import setuptools

setuptools.setup(
    name = "qbuki",
    version = "0.1.7",
    author = "Matthew Weiss",
    author_email = "heyredhat@gmail.com",
    description = "QBism for Python",
    url = "https://github.com/heyredhat/qbuki",
    packages = setuptools.find_packages(),
    include_package_data = True)